// --------------------------------------------------------------------------
//           OpenMS -- Open-Source Mass Spectrometry
// --------------------------------------------------------------------------
// Copyright The OpenMS Team -- Eberhard Karls University Tuebingen,
// ETH Zurich, and Freie Universitaet Berlin 2002-2020.
//
// This software is released under a three-clause BSD license:
//  * Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//  notice, this list of conditions and the following disclaimer in the
//  documentation and/or other materials provided with the distribution.
//  * Neither the name of any author or any participating institution
//  may be used to endorse or promote products derived from this software
//  without specific prior written permission.
// For a full list of authors, refer to the file AUTHORS.
// --------------------------------------------------------------------------
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
// INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// --------------------------------------------------------------------------
// $Maintainer: Chris Bielow $
// $Authors: Tom Waschischeck $
// --------------------------------------------------------------------------

#include <OpenMS/CONCEPT/ClassTest.h>
#include <OpenMS/test_config.h>

///////////////////////////

#include <OpenMS/QC/QCBase.h>
#include <OpenMS/KERNEL/MSExperiment.h>

///////////////////////////

using namespace OpenMS;

START_TEST(SpectraMap, "$Id$")

  QCBase::SpectraMap* ptr = nullptr;
  QCBase::SpectraMap* nulpt = nullptr;
  START_SECTION(QCBase::SpectraMap())
    {
      ptr = new QCBase::SpectraMap();
      TEST_NOT_EQUAL(ptr, nulpt)
    }
  END_SECTION

  START_SECTION(~QCBase::SpectraMap())
    {
      delete ptr;
    }
  END_SECTION
  
  MSExperiment exp;
  MSSpectrum spec1;
  spec1.setNativeID("XTandem::0");
  MSSpectrum spec2;
  spec2.setNativeID("XTandem::1");
  MSSpectrum spec3;
  spec3.setNativeID("XTandem::2");
  exp.setSpectra({spec1,spec2,spec3});
  
  START_SECTION(QCBase::SpectraMap::calculateMap(const MSExperiment& exp))
    QCBase::SpectraMap spec_map;
    spec_map.calculateMap(exp);
    ABORT_IF(spec_map.size() != 3);
    TEST_EQUAL(spec_map.at("XTandem::0"), 0);
    TEST_EQUAL(spec_map.at("XTandem::1"), 1);
    TEST_EQUAL(spec_map.at("XTandem::2"), 2);
    TEST_EXCEPTION(Exception::ElementNotFound, spec_map.at("XTandem::15"));
  END_SECTION

  START_SECTION(QCBase::SpectraMap::SpectraMap(const MSExperiment& exp))
    QCBase::SpectraMap spec_map(exp);
    TEST_EQUAL(spec_map.size(), 3);
  END_SECTION
  
  START_SECTION(QCBase::SpectraMap::empty())
    QCBase::SpectraMap spec_map;
    TEST_EQUAL(spec_map.empty(),true);
  END_SECTION

  START_SECTION(QCBase::SpectraMap::clear())
    QCBase::SpectraMap spec_map;
    spec_map.calculateMap(exp);
    TEST_EQUAL(spec_map.empty(),false);
    spec_map.clear();
    TEST_EQUAL(spec_map.empty(),true);
  END_SECTION
  
  START_SECTION(QCBase::SpectraMap::at(const String& identifier))
    NOT_TESTABLE;
  END_SECTION
  
  START_SECTION(QCBase::SpectraMap::size())
    NOT_TESTABLE;
  END_SECTION
  
END_TEST

