// --------------------------------------------------------------------------
//                   OpenMS -- Open-Source Mass Spectrometry
// --------------------------------------------------------------------------
// Copyright The OpenMS Team -- Eberhard Karls University Tuebingen,
// ETH Zurich, and Freie Universitaet Berlin 2002-2020.
//
// This software is released under a three-clause BSD license:
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//  * Neither the name of any author or any participating institution
//    may be used to endorse or promote products derived from this software
//    without specific prior written permission.
// For a full list of authors, refer to the file AUTHORS.
// --------------------------------------------------------------------------
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
// INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// --------------------------------------------------------------------------
// $Maintainer: Timo Sachsenberg $
// $Authors: Marc Sturm $
// --------------------------------------------------------------------------

// OpenMS includes
#include <OpenMS/VISUAL/DIALOGS/FeatureEditDialog.h>
#include <ui_FeatureEditDialog.h>

using namespace std;

namespace OpenMS
{

  FeatureEditDialog::FeatureEditDialog(QWidget * parent) :
    QDialog(parent),
    feature_(),
    ui_(new Ui::FeatureEditDialogTemplate)
  {
    ui_->setupUi(this);
  }

  FeatureEditDialog::~FeatureEditDialog()
  {
    delete ui_;
  }

  void FeatureEditDialog::setFeature(const Feature & feature)
  {
    //copy feature
    feature_ = feature;
    //update widgets according to feature data
    ui_->mz_->setValue(feature_.getMZ());
    ui_->rt_->setValue(feature_.getRT());
    ui_->int_->setValue(feature_.getIntensity());
    ui_->charge_->setValue(feature_.getCharge());
  }

  const Feature & FeatureEditDialog::getFeature() const
  {
    //update feature data according to widget
    feature_.setMZ(ui_->mz_->value());
    feature_.setRT(ui_->rt_->value());
    feature_.setIntensity(ui_->int_->value());
    feature_.setCharge(ui_->charge_->value());

    //return feature
    return feature_;
  }

} // namespace
