// --------------------------------------------------------------------------
//                   OpenMS -- Open-Source Mass Spectrometry
// --------------------------------------------------------------------------
// Copyright The OpenMS Team -- Eberhard Karls University Tuebingen,
// ETH Zurich, and Freie Universitaet Berlin 2002-2020.
//
// This software is released under a three-clause BSD license:
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//  * Neither the name of any author or any participating institution
//    may be used to endorse or promote products derived from this software
//    without specific prior written permission.
// For a full list of authors, refer to the file AUTHORS.
// --------------------------------------------------------------------------
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
// INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// --------------------------------------------------------------------------
// $Maintainer: Timo Sachsenberg $
// $Authors: Lukas Mueller, Markus Mueller, Peter Kunszt $
// --------------------------------------------------------------------------
//
/*
 * SuperHirnUtil.h
 *
 *  Created on: Oct 10, 2011
 *      Author: pkunszt
 */

#pragma once

#include <OpenMS/TRANSFORMATIONS/FEATUREFINDER/SUPERHIRN/SuperHirnConfig.h>

#include <OpenMS/CONCEPT/Types.h>
#include <cmath>

namespace OpenMS
{

  class SUPERHIRN_DLLAPI SuperHirnUtil
  {
public:

    /*
     * @brief Compare two masses at the PPM value and decide if they fall into the m/z tolerance window
     *
     */
    static bool compareMassValuesAtPPMLevel(double massA, double massB, double ppmTolerance);

    /*
     * @brief Get mass error at PPM value - basically ppmTolerance * m/z /10^6
     */
    static double getMassErrorAtPPMLevel(double mz, double ppmTolerance);
  };
//////////////////////////////////////////////////////
// compare two masses at the PPM value and decide
// if they fall into the m/z tolerance window
  inline bool SuperHirnUtil::compareMassValuesAtPPMLevel(double mzA, double mzB, double ppmTol)
  {

    // take the average mass, parts per million:
    double avMassPPM = (mzA + mzB) / 2000000.0;
    double ppmDeltaTol = avMassPPM * ppmTol;

    double deltaMass = fabs(mzA - mzB);
    if (deltaMass > ppmDeltaTol)
    {
      return false;
    }

    return true;
  }

//////////////////////////////////////////////////////
// get the mass error at the PPM value
  inline double SuperHirnUtil::getMassErrorAtPPMLevel(double mz, double ppmTol)
  {
    return mz * ppmTol / 1000000.00;
  }

}
