// --------------------------------------------------------------------------
//                   OpenMS -- Open-Source Mass Spectrometry
// --------------------------------------------------------------------------
// Copyright The OpenMS Team -- Eberhard Karls University Tuebingen,
// ETH Zurich, and Freie Universitaet Berlin 2002-2020.
//
// This software is released under a three-clause BSD license:
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//  * Neither the name of any author or any participating institution
//    may be used to endorse or promote products derived from this software
//    without specific prior written permission.
// For a full list of authors, refer to the file AUTHORS.
// --------------------------------------------------------------------------
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
// INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// --------------------------------------------------------------------------
// $Maintainer: Douglas McCloskey, Pasquale Domenico Colaianni $
// $Authors: Douglas McCloskey, Pasquale Domenico Colaianni $
// --------------------------------------------------------------------------

#pragma once


#include <OpenMS/ANALYSIS/QUANTITATION/AbsoluteQuantitationMethod.h>
#include <OpenMS/CONCEPT/LogStream.h>
#include <OpenMS/FORMAT/CsvFile.h>
#include <map>

namespace OpenMS
{
  /**
    @brief File adapter for AbsoluteQuantitationMethod files.

    Loads and stores .csv or .tsv files describing an AbsoluteQuantitationMethod.

    @ingroup FileIO
  */
  class OPENMS_DLLAPI AbsoluteQuantitationMethodFile :
    private CsvFile
  {
public:
    ///Default constructor
    AbsoluteQuantitationMethodFile() = default;
    ///Destructor
    ~AbsoluteQuantitationMethodFile() = default;

    /**
      @brief Loads an AbsoluteQuantitationMethod file.

      @exception Exception::FileNotFound is thrown if the file could not be opened
      @exception Exception::ParseError is thrown if an error occurs during parsing

      @param[in] filename The input file name.
      @param[out] aqm_list Output variable where the AbsoluteQuantitationMethod data is loaded.
    */
    void load(const String & filename, std::vector<AbsoluteQuantitationMethod> & aqm_list);

    /**
      @brief Stores an AbsoluteQuantitationMethod file.

      @exception Exception::UnableToCreateFile is thrown if the file could not be created

      @param[in] filename The output file name.
      @param[in] aqm_list The AbsoluteQuantitationMethod data to write into the file.
    */
    void store(const String & filename, const std::vector<AbsoluteQuantitationMethod> & aqm_list);

protected:
    /**
      @brief Parses a line into the members of AbsoluteQuantitationMethod.

      @param[in] line A line of the .csv file.
      @param[in] headers A map of header strings to column positions.
      @param[out] aqm AbsoluteQuantitationMethod.
    */
    void parseLine_(
      const StringList & line,
      const std::map<String, Size> & headers,
      AbsoluteQuantitationMethod & aqm
    ) const;

    /**
      @brief Helper method which takes care of converting the given value to the desired type,
      based on the header (here `key`) information.

      @param[in] key The header name with which the correct conversion is chosen
      @param[in] value The value to be converted
      @param[in,out] params The object where the new value is saved
    */
    void setCastValue_(const String& key, const String& value, Param& params) const;

  };

} // namespace OpenMS

