

        ## # TODO nested STL
        ## #  void computeHighestPeptides(libcpp_vector[libcpp_vector[double] ] & peptide_site_scores, libcpp_vector[ProbablePhosphoSites] & sites, libcpp_vector[libcpp_vector[Size] ] & permutations) nogil except +

        ## #Computes the site determing_ions for the given AS and sequences in candidates
        ## void compute_site_determining_ions(libcpp_vector[MSSpectrum[RichPeak1D]] & th_spectra, ProbablePhosphoSites & candidates, Int charge, libcpp_vector[MSSpectrum[RichPeak1D]] & site_determining_ions) nogil except +

        ## libcpp_vector[ size_t ] computeTupel_(AASequence & without_phospho)
        ## # TODO nested STL
        ## # libcpp_vector[ libcpp_vector[ size_t ] ] computePermutations_(libcpp_vector[ size_t ] tupel, Int number_of_phospho_sites)
