// --------------------------------------------------------------------------
//                   OpenMS -- Open-Source Mass Spectrometry
// --------------------------------------------------------------------------
// Copyright The OpenMS Team -- Eberhard Karls University Tuebingen,
// ETH Zurich, and Freie Universitaet Berlin 2002-2020.
//
// This software is released under a three-clause BSD license:
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//  * Neither the name of any author or any participating institution
//    may be used to endorse or promote products derived from this software
//    without specific prior written permission.
// For a full list of authors, refer to the file AUTHORS.
// --------------------------------------------------------------------------
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
// INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// --------------------------------------------------------------------------
// $Maintainer: Timo Sachsenberg $
// $Authors: Katharina Albers $
// --------------------------------------------------------------------------

#include <OpenMS/ANALYSIS/MAPMATCHING/MapAlignmentEvaluationAlgorithm.h>

// Derived classes are included here
#include <OpenMS/ANALYSIS/MAPMATCHING/MapAlignmentEvaluationAlgorithmPrecision.h>
#include <OpenMS/ANALYSIS/MAPMATCHING/MapAlignmentEvaluationAlgorithmRecall.h>

#include <OpenMS/CONCEPT/Factory.h>

namespace OpenMS
{
  //register products here
  void MapAlignmentEvaluationAlgorithm::registerChildren()
  {
    Factory<MapAlignmentEvaluationAlgorithm>::registerProduct(MapAlignmentEvaluationAlgorithmPrecision::getProductName(), &MapAlignmentEvaluationAlgorithmPrecision::create);
    Factory<MapAlignmentEvaluationAlgorithm>::registerProduct(MapAlignmentEvaluationAlgorithmRecall::getProductName(), &MapAlignmentEvaluationAlgorithmRecall::create);
  }

  // TODO consider using (RT,MZ,IT) as a unique identifier ?
  bool MapAlignmentEvaluationAlgorithm::isSameHandle(const FeatureHandle& lhs, const FeatureHandle& rhs, const double& rt_dev, const double& mz_dev, const Peak2D::IntensityType& int_dev, const bool use_charge)
  {
#if 1
    // use (RT,MZ,IT) as "unique" identifier
    if (fabs(lhs.getRT() - rhs.getRT()) > rt_dev)
      return false; // TODO MAGIC_ALERT

    if (fabs(lhs.getMZ() - rhs.getMZ()) > mz_dev)
      return false; // TODO MAGIC_ALERT

    if (fabs(lhs.getIntensity() - rhs.getIntensity()) > int_dev)
      return false; // TODO MAGIC_ALERT

    if (use_charge && (lhs.getCharge() != rhs.getCharge()))
      return false;

    return true;

#else
    // use (map index, element index) as unique identifier
    return lhs.getMapIndex() == rhs.getMapIndex() && lhs.getElementIndex() == rhs.getElementIndex();

#endif
  }

  MapAlignmentEvaluationAlgorithm::MapAlignmentEvaluationAlgorithm()
  {
  }

  MapAlignmentEvaluationAlgorithm::~MapAlignmentEvaluationAlgorithm()
  {
  }

}
