// --------------------------------------------------------------------------
//                   OpenMS -- Open-Source Mass Spectrometry
// --------------------------------------------------------------------------
// Copyright The OpenMS Team -- Eberhard Karls University Tuebingen,
// ETH Zurich, and Freie Universitaet Berlin 2002-2020.
//
// This software is released under a three-clause BSD license:
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//  * Neither the name of any author or any participating institution
//    may be used to endorse or promote products derived from this software
//    without specific prior written permission.
// For a full list of authors, refer to the file AUTHORS.
// --------------------------------------------------------------------------
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
// INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// --------------------------------------------------------------------------
// $Maintainer: Douglas McCloskey, Pasquale Domenico Colaianni $
// $Authors: Douglas McCloskey, Pasquale Domenico Colaianni $
// --------------------------------------------------------------------------

#pragma once

#include <OpenMS/config.h> // OPENMS_DLLAPI
#include <OpenMS/ANALYSIS/OPENSWATH/MRMFeatureSelector.h>
#include <OpenMS/KERNEL/FeatureMap.h>
#include <vector>

namespace OpenMS
{
  /**
    Class used to batch multiple calls to `MRMFeatureSelector`'s methods.
    The class offers a generic batch method (where the user is supposed to pass
    a `MRMFeatureSelector` derived object) and two specialized versions (Score and QMIP).
  */
  class OPENMS_DLLAPI MRMBatchFeatureSelector
  {
public:
    MRMBatchFeatureSelector() = delete;
    ~MRMBatchFeatureSelector() = delete;

    /**
      Calls `feature_selector.selectMRMFeature()` feeding it the parameters found in `parameters`.
      It calls said method `parameters.size()` times, using the result of each cycle as input
      for the next cycle.

      @param[in] feature_selector Base class for the feature selector to use
      @param[in] features Input features
      @param[out] selected_features Selected features
      @param[in] parameters Vector of parameters for the multiple calls to the selector
    */
    static void batchMRMFeatures(
      const MRMFeatureSelector& feature_selector,
      const FeatureMap& features,
      FeatureMap& selected_features,
      const std::vector<MRMFeatureSelector::SelectorParameters>& parameters
    );

    /// Calls `batchMRMFeatures()` using a `MRMFeatureSelectorScore` selector
    static void batchMRMFeaturesScore(
      const FeatureMap& features,
      FeatureMap& selected_features,
      const std::vector<MRMFeatureSelector::SelectorParameters>& parameters
    );

    /// Calls `batchMRMFeatures()` using a `MRMFeatureSelectorQMIP` selector
    static void batchMRMFeaturesQMIP(
      const FeatureMap& features,
      FeatureMap& selected_features,
      const std::vector<MRMFeatureSelector::SelectorParameters>& parameters
    );
  };
}
