// --------------------------------------------------------------------------
//                   OpenMS -- Open-Source Mass Spectrometry
// --------------------------------------------------------------------------
// Copyright The OpenMS Team -- Eberhard Karls University Tuebingen,
// ETH Zurich, and Freie Universitaet Berlin 2002-2020.
//
// This software is released under a three-clause BSD license:
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//  * Neither the name of any author or any participating institution
//    may be used to endorse or promote products derived from this software
//    without specific prior written permission.
// For a full list of authors, refer to the file AUTHORS.
// --------------------------------------------------------------------------
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
// INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// --------------------------------------------------------------------------
// $Maintainer: Timo Sachsenberg $
// $Authors:  Marc Sturm, Clemens Groepl $
// --------------------------------------------------------------------------

#include <OpenMS/APPLICATIONS/ParameterInformation.h>

namespace OpenMS
{
  ParameterInformation::ParameterInformation(const String& n, ParameterTypes t, const String& arg, const DataValue& def, const String& desc, bool req, bool adv, const StringList& tag_values) :
    name(n),
    type(t),
    default_value(def),
    description(desc),
    argument(arg),
    required(req),
    advanced(adv),
    tags(tag_values),
    valid_strings(),
    min_int(-std::numeric_limits<Int>::max()),
    max_int(std::numeric_limits<Int>::max()),
    min_float(-std::numeric_limits<double>::max()),
    max_float(std::numeric_limits<double>::max())
  {
  }

  ParameterInformation::ParameterInformation() :
    name(),
    type(NONE),
    default_value(),
    description(),
    argument(),
    required(true),
    advanced(false),
    tags(),
    valid_strings(),
    min_int(-std::numeric_limits<Int>::max()),
    max_int(std::numeric_limits<Int>::max()),
    min_float(-std::numeric_limits<double>::max()),
    max_float(std::numeric_limits<double>::max())
  {
  }

  ParameterInformation& ParameterInformation::operator=(const ParameterInformation& rhs)
  {
    if (&rhs == this) return *this;

    name = rhs.name;
    type = rhs.type;
    default_value = rhs.default_value;
    description = rhs.description;
    argument = rhs.argument;
    required = rhs.required;
    advanced = rhs.advanced;
    tags = rhs.tags;
    valid_strings = rhs.valid_strings;
    min_int = rhs.min_int;
    max_int = rhs.max_int;
    min_float = rhs.min_float;
    max_float = rhs.max_float;

    return *this;
  }

}
