// --------------------------------------------------------------------------
//                   OpenMS -- Open-Source Mass Spectrometry
// --------------------------------------------------------------------------
// Copyright The OpenMS Team -- Eberhard Karls University Tuebingen,
// ETH Zurich, and Freie Universitaet Berlin 2002-2020.
//
// This software is released under a three-clause BSD license:
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//  * Neither the name of any author or any participating institution
//    may be used to endorse or promote products derived from this software
//    without specific prior written permission.
// For a full list of authors, refer to the file AUTHORS.
// --------------------------------------------------------------------------
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
// INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// --------------------------------------------------------------------------
// $Maintainer: Hannes Roest $
// $Authors: Hannes Roest $
// --------------------------------------------------------------------------

#include <OpenMS/FORMAT/HDF5Connector.h>

#include <OpenMS/CONCEPT/Exception.h>

#include "H5Cpp.h"

using namespace H5;

namespace OpenMS
{

  HDF5Connector::~HDF5Connector()
  {
    close();
  }

  void HDF5Connector::close()
  {
    if (file_)
    {
      file_->flush(H5F_SCOPE_LOCAL);
      file_->close();
      delete file_;
      file_ = nullptr;
    }
  }

  HDF5Connector::HDF5Connector(const String& filename, bool createNewFile)
  {
    // H5F_ACC_TRUNC - Truncate file, if it already exists, erasing all data previously stored in the file.
    // H5F_ACC_EXCL - Fail if file already exists. H5F_ACC_TRUNC and H5F_ACC_EXCL are mutually exclusive
    // H5F_ACC_RDONLY - Open file as read-only, if it already exists, and fail, otherwise
    // H5F_ACC_RDWR - Open file for read/write, if it already exists, and fail, otherwise
    unsigned int openFlag = H5F_ACC_RDWR;
    if (createNewFile) openFlag = H5F_ACC_TRUNC;

    FileCreatPropList fcparm = FileCreatPropList::DEFAULT;
    FileAccPropList faparm = FileAccPropList::DEFAULT;
    file_ = new H5::H5File(filename, openFlag, fcparm, faparm);
  }

}


